
import java.io.*;
import java.util.regex.*;
import java.util.ArrayList;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * Created by IntelliJ IDEA.
 * User: naeem
 * Date: Mar 24, 2004
 * Time: 2:15:49 PM
 */

/**
 * The server of the one client-server file downloading program
 * This server read the first line of the file named CONFIG.cfg to assign the home directory.
 * Server waits for incoming request and according to the pattern in reqular expression returns the
 * files that matches that pattern and ask client for choosed file and then send it.
 */
public class ServerSide {

    /**
     * This function is main!
     *
     * @param args command line parameters
     */
    public static void main(String[] args) {
        try {
            new ServerSide();
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use Options | File Templates.
        }
    }

    /**
     * This is the Constructor! initing files and defining the connection for listening
     *
     * @throws IOException in case of IOError
     */
    public ServerSide() throws IOException {
        try {
            initFiles();
            System.out.println("Files has inited");
            serversocket = new ServerSocket(PORT);
            System.out.println("Server Socket initialized");
            run();
        } finally {
            serversocket.close();
        }
    }

    /**
     * Always waiting for new connection to serve it
     *
     * @throws IOException in case of IOError
     */
    private void run() throws IOException {
        while (true) {
            System.out.println("");
            System.out.println("Waiting for new connection ...");
            Object[] inout = shakeHand();

            if (inout == null)
                continue;

            System.out.println("There is incoming connection with Hello");

            ObjectInputStream in = (ObjectInputStream) inout[0];
            ObjectOutputStream out = (ObjectOutputStream) inout[1];
            String pat;

            try {
                pat = (String) in.readObject();
            } catch (ClassNotFoundException cnfe) {
                System.out.println("The read object must be String! " + cnfe.getMessage());
                socket.close();
                continue;
            } catch (ClassCastException cce) {
                System.out.println("The read object must be String! " + cce.getMessage());
                socket.close();
                continue;
            }

            System.out.println("Desired pattern is: " + pat);

            ArrayList matched = matchedFiles(pat);
            out.writeObject(new Integer(matched.size()));
            out.flush();

            for (int i = 0; i < matched.size(); i++) {
                out.writeObject(new Integer(i));
                out.writeObject(((File) matched.get(i)).getName());
                out.flush();
            }

            if( matched.size() > 0)
                sendFile(matched, in, out);

            socket.close();
            System.out.println("Connection closed");
        }
    }

    /**
     * Reads the CONFIG.cfg file and creating list of files that are in that directory.
     */
    private void initFiles() {
        try {
            BufferedReader conf = new BufferedReader(new FileReader("CONFIG.cfg"));
            path = conf.readLine();
        } catch (FileNotFoundException fnfe) {
            System.out.println("Cannot locate input file! " + fnfe.getMessage());
            System.exit(0);
        } catch (IOException ioe) {
            System.out.println("The config file is empty! " + ioe.getMessage());
            System.exit(0);
        }

        try {
            File dir = new File(path);
            myfiles = dir.listFiles();
        } catch (SecurityException se) {
            System.out.println("You have not permission to read that file!  " + se.getMessage());
            System.exit(0);
        }

        for (int i = 0; i < myfiles.length; i++) {
            if (!myfiles[i].isFile())
                myfiles[i] = null;
        }
    }

    /**
     * Getting pattern it returns the arraylist that contains the list of matched files.
     *
     * @param pat desired pattern that most be matched
     * @return matched items
     */
    private ArrayList matchedFiles(String pat) {
        ArrayList retval = new ArrayList();

        for (int i = 0; i < myfiles.length; i++)
            if (myfiles[i] != null && Pattern.matches(pat, myfiles[i].getName()))
                retval.add(myfiles[i]);

        return retval;
    }

    /**
     * Following protocol and stablishing input and output streams
     *
     * @return array that contains input(0) and output(1) streams and null in case of error
     * @throws IOException in case of IOError
     */
    private Object[] shakeHand() throws IOException {
        // Blocks until a connection occurs:
        socket = serversocket.accept();

        ObjectInputStream in = new ObjectInputStream(socket.getInputStream());
//                new BufferedInputStream( socket.getInputStream()));

        // Output is automatically flushed by PrintWriter:
        ObjectOutputStream out = new ObjectOutputStream(socket.getOutputStream());
//                new BufferedOutputStream( socket.getOutputStream()));

        String str = null;

        try {
            str = (String) in.readObject();
        } catch (ClassNotFoundException cnfe) {
            System.out.println("Problem in working with sockets! " + cnfe.getMessage());
            socket.close();
            return null;
        } catch (ClassCastException cce) {
            System.out.println("Problem in working with sockets! " + cce.getMessage());
            socket.close();
            return null;
        }

        if (!str.equals("HELLO")) {
            System.out.println("Didn't follow the protocol");
            socket.close();
            return null;
        }

        return new Object[]{in, out};
    }

    /**
     * Sending file according to choice from client
     *
     * @param matched the matched items
     * @param in input socket stream from host
     * @param out output socket stream to host
     * @throws IOException in case of IOError
     */
    private void sendFile(ArrayList matched, ObjectInputStream in, ObjectOutputStream out) throws IOException {
        int choice;

        try {
            choice = ((Integer) in.readObject()).intValue();
        } catch (ClassNotFoundException cnfe) {
            System.out.println("The read object must be Integer! " + cnfe.getMessage());
            return;
        } catch (ClassCastException cce) {
            System.out.println("The read object must be Integer! " + cce.getMessage());
            return;
        }

        if (choice == -1)
            return;

        long size = ((File) matched.get(choice)).length();
        out.writeObject(((File) matched.get(choice)).getName());
        out.writeObject( new Long(size));
        out.flush();

        System.out.println("The File Name: " + ((File) matched.get(choice)).getName());

        FileInputStream file = new FileInputStream( path + "/" + ((File) matched.get(choice)).getName());

        for( int i = 0; i < size; i++){
            int temp = file.read();
//            out.writeObject(new Integer( temp));
            out.writeByte( temp);
        }

        out.flush();
        file.close();
    }

    /**
     * Port number that server listen to it
     */
    public static final int PORT = 3131;

    /**
     * The server socket which listen to incoming connections
     */
    private ServerSocket serversocket;

    /**
     * The stablished connection
     */
    private Socket socket;

    /**
     * First line of CONFIG.cfg
     */
    private String path;

    /**
     * List of files that are in home directory
     */
    private File[] myfiles;
}
